class TextSplitter:
    def __init__(self, chunk_size=1000, chunk_overlap=200):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        if self.chunk_overlap >= self.chunk_size:
            raise ValueError('Cannot have chunkOverlap >= chunkSize')

    def split_text(self, text: str) -> list[str]:
        raise NotImplementedError

    def create_documents(self, texts: list[str]) -> list[str]:
        documents = []
        for text in texts:
            documents.extend(self.split_text(text))
        return documents

    def split_documents(self, documents: list[str]) -> list[str]:
        return self.create_documents(documents)

    def join_docs(self, docs: list[str], separator: str) -> str:
        text = separator.join(docs).strip()
        return text if text else None

    def merge_splits(self, splits: list[str], separator: str) -> list[str]:
        docs = []
        current_doc = []
        total = 0
        for d in splits:
            _len = len(d)
            if total + _len >= self.chunk_size:
                if total > self.chunk_size:
                    print(f'Created a chunk of size {total}, which is longer than the specified {self.chunk_size}')
                if current_doc:
                    doc = self.join_docs(current_doc, separator)
                    if doc:
                        docs.append(doc)
                    while total > self.chunk_overlap or (total + _len > self.chunk_size and total > 0):
                        total -= len(current_doc[0])
                        current_doc.pop(0)
            current_doc.append(d)
            total += _len
        doc = self.join_docs(current_doc, separator)
        if doc:
            docs.append(doc)
        return docs

class RecursiveCharacterTextSplitter(TextSplitter):
    def __init__(self, chunk_size=1000, chunk_overlap=200, separators=None):
        super().__init__(chunk_size, chunk_overlap)
        self.separators = separators or ['\n\n', '\n', '.', ',', '>', '<', ' ', '']

    def split_text(self, text: str) -> list[str]:
        final_chunks = []

        separator = self.separators[-1]
        for s in self.separators:
            if s == '' or s in text:
                separator = s
                break

        splits = text.split(separator) if separator else list(text)

        good_splits = []
        for s in splits:
            if len(s) < self.chunk_size:
                good_splits.append(s)
            else:
                if good_splits:
                    merged_text = self.merge_splits(good_splits, separator)
                    final_chunks.extend(merged_text)
                    good_splits = []
                final_chunks.extend(self.split_text(s))
        if good_splits:
            merged_text = self.merge_splits(good_splits, separator)
            final_chunks.extend(merged_text)
        return final_chunks


import os, asyncio
from tavily import TavilyClient
from crawl4ai import AsyncWebCrawler
from crawl4ai.extraction_strategy import JsonCssExtractionStrategy
from dotenv import load_dotenv
import html2text

# Load environment variables from .env file
load_dotenv()

TAVILY_API_KEY = os.getenv('TAVILY_API_KEY')
# Initialize TAVily client
tavily_client = TavilyClient(api_key=TAVILY_API_KEY)

# Define your search query
query = 'dog behavior studies in Taiwan'

# Asynchronous crawling function
async def crawl_url(crawler, url):
    try:
        # Crawl the URL asynchronously
        result = await crawler.arun(url=url)
        
        # Check if crawl was successful
        if result.success:
            print(f"\nCrawled {url}:")
            print(f"Content (first 200 chars): {result.html[:200]}...")
            
            # Convert HTML to Markdown
            converter = html2text.HTML2Text()
            converter.ignore_links = False  # Set to True to ignore links
            markdown_content = converter.handle(result.html)

            # Save to a Markdown file
            output_file = f"{url.replace('https://', '').replace('/', '_')}.md"
            with open(output_file, "w", encoding="utf-8") as file:
                file.write(markdown_content)

            print(f"Markdown content saved to '{output_file}'")

            # Optional: Extract structured data (e.g., paragraphs) using CSS
            extraction_strategy = JsonCssExtractionStrategy(
                "p", "paragraphs", return_type="list"
            )
            extracted_data = await crawler.aextract(url=url, extraction_strategy=extraction_strategy)
            if extracted_data and "paragraphs" in extracted_data:
                print(f"Extracted paragraphs (first 2):")
                for i, para in enumerate(extracted_data["paragraphs"][:2], 1):
                    print(f"{i}. {para[:100]}...")
        else:
            print(f"Failed to crawl {url}: {result.error_message}")
            
    except Exception as e:
        print(f"Error crawling {url}: {str(e)}")

# Main async function to run the search and crawl
async def main():
    try:
        # Perform search with TAVily API (top 5 results)
        search_results = tavily_client.search(query, max_results=5)
        
        # Extract URLs from search results
        urls = [result["url"] for result in search_results["results"]]
        print("Top URLs found:")
        for i, url in enumerate(urls, 1):
            print(f"{i}. {url}")

        # Initialize AsyncWebCrawler
        async with AsyncWebCrawler(verbose=True) as crawler:
            # Warm up the crawler (optional for performance)
            await crawler.awarmup()

            # Crawl all URLs concurrently
            tasks = [crawl_url(crawler, url) for url in urls]
            await asyncio.gather(*tasks)

    except Exception as e:
        print(f"Error during search or crawling: {str(e)}")

# Run the async main function
if __name__ == "__main__":
    asyncio.run(main())


from typing import Optional, List 


class ResearchProgress:
    def __init__(self, current_depth: int, total_depth: int, current_breadth: int, total_breadth: int, total_queries: int, completed_queries: int, current_query: Optional[str] = None):
        self.current_depth = current_depth
        self.total_depth = total_depth
        self.current_breadth = current_breadth
        self.total_breadth = total_breadth
        self.current_query = current_query
        self.total_queries = total_queries
        self.completed_queries = completed_queries

class ResearchResult:
    def __init__(self, learnings: List[str], visited_urls: List[str]):
        self.learnings = learnings
        self.visited_urls = visited_urls


import os, re
from typing import Any, Dict, List, TypedDict
import openai
from tavily import TavilyClient
from crawl4ai import AsyncWebCrawler
from firecrawl import FirecrawlApp
from dotenv import load_dotenv
import html2text

from tiktoken import get_encoding



from .text_splitter import RecursiveCharacterTextSplitter

from dotenv import load_dotenv
load_dotenv()

# Providers
openai.api_key = os.getenv('OPENAI_KEY')
openai.api_base = os.getenv('OPENAI_ENDPOINT', 'https://api.openai.com/v1')

custom_model = os.getenv('OPENAI_MODEL', 'o3-mini')

MIN_CHUNK_SIZE = 140
encoder = get_encoding('o200k_base')

TAVILY_API_KEY = os.getenv('TAVILY_API_KEY')
FIRECRAWL_KEY = os.getenv('FIRECRAWL_KEY')

# Trim prompt to maximum context size
def trim_prompt(prompt: str, context_size: int = int(os.getenv('CONTEXT_SIZE', 128000))) -> str:
    if not prompt:
        return ''

    length = len(encoder.encode(prompt))
    if length <= context_size:
        return prompt

    overflow_tokens = length - context_size
    chunk_size = len(prompt) - overflow_tokens * 3
    if chunk_size < MIN_CHUNK_SIZE:
        return prompt[:MIN_CHUNK_SIZE]

    splitter = RecursiveCharacterTextSplitter(chunk_size=chunk_size, chunk_overlap=0)
    trimmed_prompt = splitter.split_text(prompt)[0] if splitter.split_text(prompt) else ''

    if len(trimmed_prompt) == len(prompt):
        return trim_prompt(prompt[:chunk_size], context_size)

    return trim_prompt(trimmed_prompt, context_size)


# Define Search top few number of urls by tavily_client
class TavilySearch:
    def __init__(self):
        self.name = 'tavily.TavilySearch'
        self.tavily_client = TavilyClient(api_key=TAVILY_API_KEY)

    def search(self, query: str, max_results: int) -> List[Dict[str, Any]]:
        # Perform search with TAVily API (top results)
        # Extract the string between double quotes
        match = re.search(r'"(.*?)"', query)
        if match:
            query = match.group(1)
        else:
            query = re.sub(r'[*-]', '', query)
        
        search_results = self.tavily_client.search(query, max_results=max_results)
        
        # Extract URLs from search results and filter out google.com results
        urls = [{'url': result["url"], 'title': result["title"]} for result in search_results["results"] if "google.com" not in result["url"]]
        return urls    

# Define WebCrawlerApp
class WebCrawlerApp:
    def __init__(self):
        self.name = 'crawl4ai.AsyncWebCrawler'

    async def crawl_url(self, url: str) -> str:
        try:
            crawler = AsyncWebCrawler(verbose=True)
            # Crawl the URL asynchronously
            result = await crawler.arun(url=url)
            
            # Check if crawl was successful
            if result.success:
                print(f"\nCrawled {url}:")
                print(f"Content (first 200 chars): {result.html[:200]}...")
                
                # Convert HTML to Markdown
                converter = html2text.HTML2Text()
                converter.ignore_links = False  # Set to True to ignore links
                markdown_content = converter.handle(result.html)
                return markdown_content
            else:
                print(f"Failed to crawl {url}: {result.error_message}")
                
        except Exception as e:
            print(f"Error crawling {url}: {str(e)}")
    

# Define WebFirecrawlApp
class WebFirecrawlApp:
    def __init__(self):
        self.api_key = FIRECRAWL_KEY
        #self.api_url = config.get('apiUrl')

    def search(self,query: str, max_results: int) -> List[Dict[str, Any]]:
        try:
            app = FirecrawlApp(api_key=self.api_key)
            # Search for the query
            search_result = app.search(query)
            # Extract URLs from the search result
            urls = [item['url'] for item in search_result['data']]
            return {'urls': urls[:max_results]}
            
        except Exception as e:
            print(f"Error searching for {query}: {str(e)}")

    async def crawl_url(self, url: str) -> str:
        try:
            app = FirecrawlApp(api_key=self.api_key)
            # Scrape the URL
            scrape_result = app.scrape_url(url, params={'formats': ['markdown']})
            return scrape_result['markdown']
            
        except Exception as e:
            print(f"Error scraping {url}: {str(e)}")
        


# Define SearchResponse
class SearchResponseItem(TypedDict):
    markdown: str
    url: str

class SearchResponse(TypedDict):
    data: List[SearchResponseItem]


import unittest
from .text_splitter import RecursiveCharacterTextSplitter

class TestRecursiveCharacterTextSplitter(unittest.TestCase):
    def setUp(self):
        self.splitter = RecursiveCharacterTextSplitter(chunk_size=50, chunk_overlap=10)

    def test_split_text_by_separators(self):
        text = 'Hello world, this is a test of the recursive text splitter.'

        self.assertEqual(
            self.splitter.split_text(text),
            ['Hello world', 'this is a test of the recursive text splitter']
        )

        self.splitter.chunk_size = 100
        self.assertEqual(
            self.splitter.split_text(
                'Hello world, this is a test of the recursive text splitter. If I have a period, it should split along the period.'
            ),
            [
                'Hello world, this is a test of the recursive text splitter',
                'If I have a period, it should split along the period.'
            ]
        )

        self.splitter.chunk_size = 110
        self.assertEqual(
            self.splitter.split_text(
                'Hello world, this is a test of the recursive text splitter. If I have a period, it should split along the period.\nOr, if there is a new line, it should prioritize splitting on new lines instead.'
            ),
            [
                'Hello world, this is a test of the recursive text splitter',
                'If I have a period, it should split along the period.',
                'Or, if there is a new line, it should prioritize splitting on new lines instead.'
            ]
        )

    def test_handle_empty_string(self):
        self.assertEqual(self.splitter.split_text(''), [])

    def test_handle_special_characters_and_large_texts(self):
        large_text = 'A' * 1000
        self.splitter.chunk_size = 200
        self.assertEqual(
            self.splitter.split_text(large_text),
            ['A' * 200] * 5
        )

        special_char_text = 'Hello!@# world$%^ &*( this) is+ a-test'
        self.assertEqual(
            self.splitter.split_text(special_char_text),
            ['Hello!@#', 'world$%^', '&*( this)', 'is+', 'a-test']
        )

    def test_handle_chunk_size_equal_to_chunk_overlap(self):
        self.splitter.chunk_size = 50
        self.splitter.chunk_overlap = 50
        with self.assertRaises(ValueError):
            self.splitter.split_text('Invalid configuration')

if __name__ == '__main__':
    unittest.main()


import asyncio
from typing import Any
from .deep_research import deep_research, write_final_report
from .feedback import generate_feedback
from .output_manager import OutputManager

output = OutputManager()

def log(*args: Any) -> None:
    output.log(*args)

async def ask_question(query: str) -> str:
    return input(query)

async def run() -> None:
    initial_query = await ask_question('What would you like to research? ')
    breadth = int(await ask_question('Enter research breadth (recommended 2-10, default 4): ') or 4)
    depth = int(await ask_question('Enter research depth (recommended 1-5, default 2): ') or 2)

    log('Creating research plan...')
    follow_up_questions = await generate_feedback(query=initial_query)  # Await the coroutine
    log('\nTo better understand your research needs, please answer these follow-up questions:')

    answers = []
    for question in follow_up_questions:
        answer = await ask_question(f'\n{question}\nYour answer: ')
        answers.append(answer)

    combined_query = f"Initial Query: {initial_query}\nFollow-up Questions and Answers:\n" + \
                     "\n".join([f"Q: {q}\nA: {answers[i]}" for i, q in enumerate(follow_up_questions)])
    log('\nResearching your topic...')
    log('\nStarting research with progress tracking...\n')

    result = await deep_research(query=combined_query, breadth=breadth, depth=depth, on_progress=output.update_progress)
    log(f"\n\nLearnings:\n\n{''.join(result.learnings)}")
    log(f"\n\nVisited URLs ({len(result.visited_urls)}):\n\n{''.join(result.visited_urls)}")
    log('Writing final report...')

    report = await write_final_report(prompt=combined_query, learnings=result.learnings, visited_urls=result.visited_urls)
    with open('output.md', 'w', encoding='utf-8') as f:
        f.write(report)

    print(f"\n\nFinal Report:\n\n{report}")
    print('\nReport has been saved to output.md')

def main():
    asyncio.run(run())

if __name__ == '__main__':
    asyncio.run(run())